'''
exceptions.py: exceptions defined by Foliage

Authors
-------

Michael Hucka <mhucka@caltech.edu> -- Caltech Library

Copyright
---------

Copyright (c) 2021-2022 by the California Institute of Technology.  This code is
open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''


# Base class.
# .............................................................................
# The base class makes it possible to use a single test to distinguish between
# exceptions generated by Foliage and exceptions generated by something else.

class FoliageException(Exception):
    '''Base class for Foliage exceptions.'''
    pass


# Exception classes.
# .............................................................................

class FolioOpFailed(FoliageException):
    '''Requested operation was unsuccessful.'''
    pass


class FolioError(FoliageException):
    '''Unrecoverable problem involving interactions with the FOLIO server.'''
    pass



'''
delete_tab.py: implementation of the "Delete records" tab

Copyright
---------

Copyright (c) 2021-2022 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

from   commonpy.data_utils import unique, pluralized, flattened
from   commonpy.exceptions import Interrupted
from   commonpy.file_utils import exists, readable
from   commonpy.interrupt import wait, reset_interrupts, interrupt, interrupted
from   decouple import config
import json
from   pywebio.input import input, select, checkbox, radio
from   pywebio.input import NUMBER, TEXT, input_update, input_group
from   pywebio.output import put_text, put_markdown, put_row, put_html
from   pywebio.output import toast, popup, close_popup, put_buttons, put_button, put_error
from   pywebio.output import use_scope, set_scope, clear, remove, put_warning
from   pywebio.output import put_success, put_info, put_table, put_grid, span
from   pywebio.output import put_tabs, put_image, put_scrollable, put_code, put_link
from   pywebio.output import put_processbar, set_processbar, put_loading
from   pywebio.output import put_column, put_scope, clear_scope, put_loading
from   pywebio.pin import pin, pin_wait_change, put_input, put_actions
from   pywebio.pin import put_textarea, put_radio, put_checkbox, put_select
from   sidetrack import set_debug, log

from   foliage.base_tab import FoliageTab
from   foliage.exceptions import *
from   foliage.export import export_data
from   foliage.folio import Folio, RecordKind, IdKind, TypeKind, Record
from   foliage.folio import unique_identifiers, back_up_record
from   foliage.ui import confirm, notify, user_file, stop_processbar
from   foliage.ui import tell_success, tell_warning, tell_failure
from   foliage.ui import note_info, note_warn, note_error
from   foliage.ui import PROGRESS_BOX, PROGRESS_TEXT


# Tab definition class.
# .............................................................................

class DeleteTab(FoliageTab):
    def contents(self):
        return {'title': 'Delete records', 'content': tab_contents()}

    def pin_watchers(self):
        return {}


# Tab creation function.
# .............................................................................

def tab_contents():
    log(f'generating delete tab contents')
    return [
        put_grid([[
            put_markdown('Input one or more item, holdings, or instance'
                         + ' identifiers (in the form of id\'s, barcodes, hrid\'s,'
                         + ' or accession numbers), or upload a file containing'
                         + ' the identifiers. **Warning**: Deleting holdings'
                         + ' **will delete all their items**, and deleting'
                         + ' instances **will delete all their holdings _and_ items**.'),
            put_button('Upload', outline = True,
                       onclick = lambda: load_file()).style('text-align: right'),
        ]], cell_widths = 'auto 100px'),
        put_textarea('textbox_delete', rows = 4),
        put_row([
            put_button('Delete records', color = 'danger',
                       onclick = lambda: do_delete()),
            put_button('Clear', outline = True,
                       onclick = lambda: clear_tab()).style('text-align: right')
        ])
    ]


# Implementation of tab functionality.
# .............................................................................

def clear_tab():
    log(f'clearing tab')
    clear('output')
    pin.textbox_delete = ''


def load_file():
    log(f'user requesting file upload')
    if (file := user_file('Upload a file containing identifiers')):
        pin.textbox_delete = file


def stop():
    log(f'stopping')
    interrupt()
    stop_processbar()


_results = []

def clear_results():
    global _results
    _results = []


def record_result(record_or_id, success, notes):
    global _results
    id = record_or_id if isinstance(record_or_id, str) else record_or_id.id
    rec = record_or_id if isinstance(record_or_id, Record) else None
    _results.append({'id': id, 'success': success, 'notes': notes, 'record': rec})


def succeeded(record_or_id, msg, why = ''):
    comment = (' (' + why + ')') if why else ''
    record_result(record_or_id, True, msg + comment)
    tell_success('Success: ' + msg + comment + '.')


def failed(record_or_id, msg, why = ''):
    comment = (' (' + why + ')') if why else ''
    record_result(record_or_id, False, msg + comment)
    id = record_or_id if isinstance(record_or_id, str) else record_or_id.id
    tell_failure(f'Failed to delete **{id}**{comment}: ' + msg + '.')


def skipped(record_or_id, msg, why = ''):
    comment = (' (' + why + ')') if why else ''
    record_result(record_or_id, False, msg + comment)
    id = record_or_id if isinstance(record_or_id, str) else record_or_id.id
    tell_warning(f'Skipped **{id}**{comment}: ' + msg + '.')


def do_delete():
    log(f'do_delete invoked')
    identifiers = unique_identifiers(pin.textbox_delete)
    if not identifiers:
        note_error('Please input at least one barcode or other type of id.')
        return
    if not confirm('Warning: if the deletions include holdings and/or instance'
                   ' records, all their associated items and holdings will'
                   ' also be deleted. Only do this if you have verified the'
                   ' implications first. Proceed?', danger = True):
        log(f'user declined to proceed')
        return
    clear_results()
    reset_interrupts()
    steps = 2*len(identifiers)       # Count getting records, for more action.
    folio = Folio()
    with use_scope('output', clear = True):
        put_grid([[
            put_scope('current_activity', [
                put_markdown('_Getting records ..._').style(PROGRESS_TEXT)]),
        ], [
            put_processbar('bar', init = 0/steps).style('margin-top: 11px'),
            put_button('Stop', outline = True, color = 'danger',
                       onclick = lambda: stop()).style('text-align: right')
        ]], cell_widths = '85% 15%').style(PROGRESS_BOX)
        try:
            done = 0
            for id in identifiers:
                with use_scope('current_activity', clear = True):
                    put_markdown(f'_Looking up {id} ..._').style(PROGRESS_TEXT)
                record = folio.record(id)
                if not record:
                    failed(id, f'unrecognized identifier **{id}**')
                    continue
                done += 1
                set_processbar('bar', done/steps)
                if record.kind not in _HANDLERS.keys():
                    skipped(id, f'deleting {record.kind} records is not supported')
                    continue
                with use_scope('current_activity', clear = True):
                    text = '_Deleting '
                    if record.kind is RecordKind.ITEM:
                        text += f'item {id} ..._'
                    elif record.kind is RecordKind.HOLDINGS:
                        text += f'holdings {id} and associated items ..._'
                    elif record.kind is RecordKind.INSTANCE:
                        text += f'instance {id} and associated holdings and items ..._'
                    elif record.kind is RecordKind.USER:
                        text += f'user {id} and associated holdings and items ..._'
                    put_markdown(text).style(PROGRESS_TEXT)
                _HANDLERS[record.kind](record)
                done += 1
                set_processbar('bar', done/steps)
            set_processbar('bar', 1)
            clear_scope('current_activity')
        except Interrupted as ex:
            tell_warning('**Stopped**.')
            return
        except Exception as ex:
            import traceback
            log('Exception info: ' + str(ex) + '\n' + traceback.format_exc())
            tell_failure(f'Error: ' + str(ex))
            return
        finally:
            stop_processbar()

        put_grid([[
            put_markdown('Finished deletions.').style('margin-top: 6px'),
            put_button('Export summary', outline = True,
                       onclick = lambda: do_export('foliage-deletions.csv'),
                       ).style('text-align: right')
        ]]).style('margin: 1.5em 17px auto 17px')


def delete(record, for_id = None):
    '''Generic low-level function to delete the given record.'''
    why = ('for request to delete ' + for_id) if for_id else ''
    if config('DEMO_MODE', cast = bool):
        log(f'demo mode in effect – pretending to delete {record.id}')
    else:
        try:
            back_up_record(record)
            folio = Folio()
            folio.delete_record(record)
        except FolioOpFailed as ex:
            failed(record, str(ex), why)
            return False
    succeeded(record, f'deleted {record.kind} record **{record.id}**', why)
    return True


def delete_holdings(holdings, for_id = None):
    '''Delete the given holdings record and all its items.'''
    # Deletions on holdings are not recursive. The items have to be deleted 1st
    # and FOLIO will give a 400 error if you try to delete a holdings record
    # while there's still an item somewhere pointing to it.
    why = 'deleting item attached to holdings record {holdings.id}'
    folio = Folio()
    # Start at the bottom: delete its items 1st.
    for item in folio.related_records(holdings.id, IdKind.HOLDINGS_ID, RecordKind.ITEM):
        if not delete(item, for_id = holdings.id):
            failed(holdings, f'unable to delete item {item.id} – stopping')
            return False
    # If we get this far, delete the holdings record.
    return delete(holdings, for_id)


def delete_instance(instance, for_id = None):
    '''Delete the given instance record.'''
    # Deletions on instances are not recursive regardless of which API you use.
    # The following is based on Kyle Banerjee's script dated 2021-11-11 at
    # https://github.com/FOLIO-FSE/shell-utilities/blob/master/instance-delete
    # but whereas that script uses the instance-storage API endpoint, this uses
    # the inventory API. (I'm told the latter will simply forward the request
    # to the instance storage API, but I decided to use the inventory API in
    # case it does more in the future.) Note that this also uses the source
    # storage API, which is a 3rd API besides the inventory API and storage
    # API used elsewhere in Foliage. (That part comes from Banerjee's script.)

    # Start by using delete_holdings(), which will delete items too.
    why = 'deleting all records attached to instance record {instance.id}'
    folio = Folio()
    for holdings in folio.related_records(instance.id, IdKind.INSTANCE_ID,
                                          RecordKind.HOLDINGS):
        if not delete_holdings(holdings, for_id = instance.id):
            failed(instance, f'unable to delete all holdings records')
            return False

    # Next, get the matched id from source storage & delete the instance there.
    def response_converter(response):
        if not response or not response.text:
            log('no response received from FOLIO')
            return None
        return json.loads(response.text)

    folio = Folio()
    srsget = f'/source-storage/records/{instance.id}/formatted?idType=INSTANCE'
    data_json = folio.request(srsget, converter = response_converter)
    if not data_json:
        failed(instance, 'unable to retrieve instance data from FOLIO SRS')
        return
    elif 'matchedId' not in data_json:
        failed(instance, 'unexpected data from FOLIO SRS – please report this')
        return
    srs_id = data_json["matchedId"]
    log(f'SRS id for {instance.id} is {srs_id}')
    srsdel = f'/source-storage/records/{srs_id}'
    folio.request(srsdel)

    # If we didn't get an exception, finally delete the instance from inventory.
    return delete(instance, for_id)


def delete_user(record, for_id = None):
    '''Delete the given user record.'''
    failed(record, 'user records cannot be deleted at this time')


_location_map = None

def init_location_map():
    global _location_map
    if _location_map is None:
        folio = Folio()
        _location_map = {x.data['id']:x.data['name']
                         for x in folio.types(TypeKind.LOCATION)}


def location(location_data):
    global _location_map
    if isinstance(location_data, dict):
        if 'name' in location_data:
            return f'{location_data["name"]} ({location_data["id"]})'
        else:
            return location_data["id"]
    elif location_data and location_data in _location_map:
        return f'{_location_map[location_data]} ({location_data})'
    return '(unknown location)'


def do_export(file_name):
    global _results
    init_location_map()
    # Output fields requested
    #   id
    #   success
    #   notes
    #   record hrid
    #   record barcode (if present)
    #   record location
    #   record title
    #   record publication year
    #   record publisher

    # People want to see titles, dates, locations. We need to do lookups to
    # get them.
    values = []
    folio = Folio()
    for result in _results:
        entry = {'Record ID'          : result['id'],
                 'Operation success'  : result['success'],
                 'Notes'              : result['notes'],
                 'Effective location' : '',
                 'Permanent location' : ''}
        if result['record']:
            rec = result['record']
            if rec.kind == RecordKind.ITEM:
                entry['Permanent location'] = location(rec.data['permanentLocationId'])
                if 'effectiveLocationId' in rec.data:
                    entry['Effective location'] = location(rec.data['effectiveLocationId'])
        values.append(entry)
    export_data(values, file_name)


_HANDLERS = {
    RecordKind.ITEM     : delete,       # The generic function suffices.
    RecordKind.HOLDINGS : delete_holdings,
    RecordKind.INSTANCE : delete_instance,
    RecordKind.USER     : delete_user,
}



'''
list_tab.py: implementation of the "List UUIDs" tab

Copyright
---------

Copyright (c) 2021-2022 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

from   commonpy.data_utils import unique, pluralized, flattened
from   commonpy.file_utils import exists, readable
from   commonpy.interrupt import wait, interrupt, interrupted, reset_interrupts
from   pprint import pformat
import pyperclip
from   pywebio.input import input, select, checkbox, radio
from   pywebio.input import NUMBER, TEXT, input_update, input_group
from   pywebio.output import put_text, put_markdown, put_row, put_html
from   pywebio.output import toast, popup, close_popup, put_buttons, put_button, put_error
from   pywebio.output import use_scope, set_scope, clear, remove, put_warning
from   pywebio.output import put_success, put_info, put_table, put_grid, span
from   pywebio.output import put_tabs, put_image, put_scrollable, put_code, put_link
from   pywebio.output import put_processbar, set_processbar, put_loading
from   pywebio.output import put_column
from   pywebio.pin import pin, pin_wait_change, put_input, put_actions
from   pywebio.pin import put_textarea, put_radio, put_checkbox, put_select
from   sidetrack import set_debug, log
import threading

from   foliage.base_tab import FoliageTab
from   foliage.export import export_records
from   foliage.folio import Folio, RecordKind, IdKind, TypeKind
from   foliage.ui import confirm, notify, stop_processbar
from   foliage.ui import note_info, note_warn, note_error, tell_success, tell_failure


# Tab definition class.
# .............................................................................

class ListTab(FoliageTab):
    def contents(self):
        return {'title': 'List UUIDs', 'content': tab_contents()}

    def pin_watchers(self):
        return {}


# Tab body.
# .............................................................................

def tab_contents():
    log(f'generating list tab contents')
    return [
        put_grid([[
            put_markdown('Select a FOLIO type to list:').style('margin-top: 6px'),
            put_select('list_type', options = [
                {'label': 'Acquisition units', 'value': TypeKind.ACQUISITION_UNIT},
                {'label': 'Address types', 'value': TypeKind.ADDRESS},
                {'label': 'Alternative title types', 'value': TypeKind.ALT_TITLE},
                {'label': 'Call number types', 'value': TypeKind.CALL_NUMBER},
                {'label': 'Classification types', 'value': TypeKind.CLASSIFICATION},
                {'label': 'Contributor types', 'value': TypeKind.CONTRIBUTOR},
                {'label': 'Contributor name types', 'value': TypeKind.CONTRIBUTOR_NAME},
                {'label': 'Department types', 'value': TypeKind.DEPARTMENT},
                {'label': 'Expense classes', 'value': TypeKind.EXPENSE_CLASS},
                {'label': 'Fixed due date schedules', 'value': TypeKind.FIXED_DUE_DATE_SCHED},
                {'label': 'Patron group types', 'value': TypeKind.GROUP},
                {'label': 'Holdings types', 'value': TypeKind.HOLDINGS},
                {'label': 'Holdings note types', 'value': TypeKind.HOLDINGS_NOTE},
                {'label': 'Holdings source types', 'value': TypeKind.HOLDINGS_SOURCE},
                {'label': 'Identifier types', 'value': TypeKind.ID},
                {'label': 'ILL policy types', 'value': TypeKind.ILL_POLICY},
                {'label': 'Instance types', 'value': TypeKind.INSTANCE},
                {'label': 'Instance format types', 'value': TypeKind.INSTANCE_FORMAT},
                {'label': 'Instance note types', 'value': TypeKind.INSTANCE_NOTE},
                {'label': 'Instance relationship types', 'value': TypeKind.INSTANCE_REL},
                {'label': 'Instance status types', 'value': TypeKind.INSTANCE_STATUS},
                {'label': 'Item note types', 'value': TypeKind.ITEM_NOTE},
                {'label': 'Item damaged status types', 'value': TypeKind.ITEM_DAMAGED_STATUS},
                {'label': 'Loan types', 'value': TypeKind.LOAN},
                {'label': 'Loan policy types', 'value': TypeKind.LOAN_POLICY},
                {'label': 'Location types', 'value': TypeKind.LOCATION},
                {'label': 'Material types', 'value': TypeKind.MATERIAL},
                {'label': 'Nature of content term types', 'value': TypeKind.NATURE_OF_CONTENT},
#                {'label': 'Order lines', 'value': TypeKind.ORDER_LINE},
                {'label': 'Organizations', 'value': TypeKind.ORGANIZATION},
                {'label': 'Service point types', 'value': TypeKind.SERVICE_POINT},
                {'label': 'Shelf location types', 'value': TypeKind.SHELF_LOCATION},
                {'label': 'Statistical code types', 'value': TypeKind.STATISTICAL_CODE},
            ]).style('margin-left: 10px; margin-bottom: 0'),
            put_button('Get list', onclick = lambda: do_list(),
                       ).style('margin-left: 10px; text-align: left'),
            put_button('Clear', outline = True, onclick = lambda: clear_tab()
                       ).style('margin-left: 10px; text-align: right'),
        ]])
    ]


# Tab implementation.
# .............................................................................

def clear_tab():
    log(f'clearing tab')
    clear('output')


def do_list():
    folio = Folio()
    reset_interrupts()
    with use_scope('output', clear = True):
        put_processbar('bar', init = 1/2)
        requested = pin.list_type
        log(f'getting list of {requested} types')
        try:
            types = folio.types(requested)
        except Exception as ex:
            import traceback
            log('Exception info: ' + str(ex) + '\n' + traceback.format_exc())
            put_html('<br>')
            tell_failure('Error: ' + str(ex))
            return
        finally:
            set_processbar('bar', 2/2)
        cleaned_name = requested.split('/')[-1].replace("-", " ")
        put_row([
            put_markdown(f'Found {len(types)} values for {cleaned_name}:'
                         ).style('margin-left: 17px; margin-top: 6px'),
            put_button('Export', outline = True,
                       onclick = lambda: export_records(types, requested),
                       ).style('text-align: right; margin-right: 17px'),
        ]).style('margin-top: 15px; margin-bottom: 14px')
        rows = []
        for item in types:
            name = item.data[TypeKind.name_key(requested)]
            title = f'Data for {cleaned_name} value "{name.title()}"'
            rows.append([name, link_button(name, item.id, title, requested),
                         copy_button(item.id).style('padding: 0; margin-right: 13px')])

        contents = [[put_markdown('**Name**'), put_markdown('**Id**'), put_text('')]]
        contents += sorted(rows, key = lambda x: x[0])
        put_grid(contents, cell_widths = 'auto auto 106px')
        stop_processbar()


def show_record(title, id, record_type):
    folio = Folio()
    try:
        log(f'getting {record_type} record {id} from FOLIO')
        recs  = folio.types(record_type)
    except Exception as ex:
        note_error(str(ex))
        return

    event = threading.Event()

    def clk(val):
        event.set()

    record  = recs[0] if isinstance(recs, list) and len(recs) > 0 else recs
    pins  = [
        put_scrollable(put_code(pformat(record.data, indent = 2)), height = 400),
        put_buttons([{'label': 'Close', 'value': 1}], onclick = clk).style('float: right'),
    ]
    popup(title = title, content = pins, size = 'large')

    event.wait()
    close_popup()


def link_button(name, id, title, record_type):
    return put_button(id, link_style = True,
                      onclick = lambda: show_record(title, id, record_type),
                      ).style('margin-left: 0; margin-top: 0.25em; margin-bottom: 0.5em')


def copy_button(text):
    def copy_to_clipboard():
        log(f'copying {text} to clipboard')
        pyperclip.copy(text)

    return put_button('Copy id', onclick = lambda: copy_to_clipboard(),
                      outline = True, small = True).style('text-align: center')



'''
lookup_tab.py: implementation of the "Look up records" tab

Copyright
---------

Copyright (c) 2021-2022 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

from   commonpy.data_utils import unique, pluralized, flattened
from   commonpy.exceptions import Interrupted
from   commonpy.file_utils import exists, readable
from   commonpy.interrupt import wait, interrupt, interrupted, reset_interrupts
import json
from   pprint import pformat
from   pywebio.input import input, select, checkbox, radio
from   pywebio.input import NUMBER, TEXT, input_update, input_group
from   pywebio.output import put_text, put_markdown, put_row, put_html
from   pywebio.output import toast, popup, close_popup, put_buttons, put_button, put_error
from   pywebio.output import use_scope, set_scope, clear, remove, put_warning
from   pywebio.output import put_success, put_info, put_table, put_grid, span
from   pywebio.output import put_tabs, put_image, put_scrollable, put_code, put_link
from   pywebio.output import put_processbar, set_processbar, put_loading
from   pywebio.output import put_column, put_scope, clear_scope, get_scope
from   pywebio.pin import pin, pin_wait_change, put_input, put_actions
from   pywebio.pin import put_textarea, put_radio, put_checkbox, put_select
from   pywebio.session import run_js, eval_js
from   sidetrack import set_debug, log
import threading

from   foliage.base_tab import FoliageTab
from   foliage.export import export_records
from   foliage.folio import Folio, RecordKind, IdKind, TypeKind
from   foliage.folio import unique_identifiers
from   foliage.ui import confirm, notify, user_file, stop_processbar
from   foliage.ui import tell_success, tell_warning, tell_failure
from   foliage.ui import note_info, note_warn, note_error, PROGRESS_BOX


# Tab definition class.
# .............................................................................

class LookupTab(FoliageTab):
    def contents(self):
        return {'title': 'Look up records', 'content': tab_contents()}

    def pin_watchers(self):
        return {}


# Tab layout.
# .............................................................................

def tab_contents():
    log(f'generating lookup tab contents')
    return [
        put_grid([[
            put_markdown('Input one or more item barcode, item id, item hrid,'
                         + ' instance id, instance hrid, instance accession'
                         + ' number, loan id, loan hrid, user id, or user'
                         + ' barcode below, or upload a file containing them.'),
            put_button('Upload', outline = True,
                       onclick = lambda: load_file()).style('text-align: right'),
        ]], cell_widths = 'auto 100px'),
        put_textarea('textbox_find', rows = 4),
        put_grid([[
            put_radio('select_kind', inline = True,
                      label = 'Kind of record to retrieve:',
                      options = [ ('Item', RecordKind.ITEM, True),
                                  ('Holdings', RecordKind.HOLDINGS),
                                  ('Instance', RecordKind.INSTANCE),
                                  ('Loan', RecordKind.LOAN),
                                  ('User', RecordKind.USER)]),
            put_checkbox("open_loans", inline = True,
                         options = [('Search open loans only', True, True)],
                         help_text = ('When searching for loans (and users,'
                                      + ' based on loans), limit searches to'
                                      + ' open loans only. Deselect'
                                      + ' to search all loans.')),
        ]], cell_widths = '54% 46%'),
        put_grid([[
            put_grid([[
                put_text('Format in which to display records:'),
            ],[
                put_radio('show_raw', inline = True,
                          options = [('Summary', 'summary', True),
                                     ('Raw data', 'json')]),
            ]]),
            put_checkbox("inventory_api", inline = True,
                         options = [('Use inventory API for items and instances',
                                     True, True)],
                         help_text = ("FOLIO's Inventory API shows more fields but"
                                      + ' some values are computed. Deselect to'
                                      + ' get pure records from the storage API.')),
        ]], cell_widths = '54% 46%'),
        put_row([
            put_button('Look up records', onclick = lambda: do_find()),
            put_text(''),    # Adds a column, pushing next item to the right.
            put_button('Clear', outline = True,
                       onclick = lambda: clear_tab()).style('text-align: right')
        ])
    ]


# Tab implementation.
# .............................................................................

_interrupted = False
_running = False
_last_textbox = ''
_last_results = {}
_last_kind = None
_last_inventory_api = True
_last_open_loans = True
_location_map = None


def load_file():
    log(f'user requesting file upload')
    if (contents := user_file('Upload a file containing identifiers')):
        pin.textbox_find = contents


def init_location_map():
    global _location_map
    if _location_map is None:
        folio = Folio()
        _location_map = {x.data['id']:x.data['name']
                         for x in folio.types(TypeKind.LOCATION)}


def inputs_are_unchanged():
    global _last_textbox
    global _last_kind
    global _last_inventory_api
    global _last_open_loans
    unchanged = (pin.textbox_find == _last_textbox
                 and pin.select_kind == _last_kind
                 and pin.inventory_api == _last_inventory_api
                 and pin.open_loans == _last_open_loans)
    log(f'field values are considered {"unchanged" if unchanged else "changed"}')
    return unchanged


def clear_tab():
    global _last_textbox
    global _last_inventory_api
    log(f'clearing tab')
    clear('output')
    pin.textbox_find = ''
    pin.inventory_api = [True]
    _last_textbox = ''
    _last_inventory_api = [True]
    _last_open_loans = [True]


def stop():
    '''Stop an ongoing lookup by setting the _interrupted flag.'''
    global _interrupted
    global _last_textbox
    log(f'stopping')
    _interrupted = True
    _last_textbox = ''
    stop_processbar()
    interrupt()
    with use_scope('output'):
        tell_warning('**Stopping** ...')


def reset():
    # Reset to state where we can run new operations.
    global _interrupted
    _interrupted = False
    reset_interrupts()


def enable_lookup_button(state):
    '''Enable the "look up records" button if True, disable if False.'''
    action = 'removeClass' if state else 'addClass'
    eval_js(f'''$("button:contains('Look up records')").{action}("disabled-button");''')


def wait_if_running():
    '''Check if the run state is running; if it is, wait until it changes.'''
    # The _running variable is set by do_find() when it's finished, and it's
    # set even if it gets interrupted. This is what we cue from.
    global _running
    if not _running:
        return
    enable_lookup_button(False)
    stop()
    # Wait in case an ongoing lookup is running, but don't wait forever.
    wait_count = 10
    while _running and wait_count > 0:
        # If the user clicks multiple times rapidly, the exception raised for
        # interrupts starts to cascade. Wrap with a try-except to avoid this.
        try:
            wait(1)
            wait_count -= 1
        except Interrupted:
            continue
    enable_lookup_button(True)


# Summary of the basic flow of control:
#
# User clicks "look up records", thus invoking do_find().
# We show progress bar & stop button while lookup is running.
# Possible scenarios:
#   1) process finishes normally
#   2) user clicks stop button
#   3) user clicks "look up records" button while lookup is running

def do_find():
    global _last_results
    global _last_textbox
    global _last_kind
    global _last_inventory_api
    global _last_open_loans
    global _location_map
    global _interrupted
    global _running
    log('do_find invoked')
    wait_if_running()
    reset()
    # Normally we'd want to find out if they input any identifiers, but I want
    # to detect *any* change to the input box, so this is a lower-level test.
    if not pin.textbox_find.strip():
        note_error('Please input at least one barcode or other id.')
        return
    identifiers = unique_identifiers(pin.textbox_find)
    if not identifiers:
        note_error('The input does not appear to contain FOLIO identifiers.')
        return
    reuse_results = False
    if inputs_are_unchanged() and user_wants_reuse():
        reuse_results = True
    else:
        _last_results = {}
    _last_textbox = pin.textbox_find
    _last_kind = pin.select_kind
    _last_inventory_api = pin.inventory_api
    _last_open_loans = pin.open_loans
    kind_wanted = pin.select_kind
    steps = len(identifiers) + 1
    folio = Folio()
    init_location_map()
    total_found = 0
    with use_scope('output', clear = True):
        put_grid([[
            put_scope('current_activity', [
                put_markdown(f'_Certain lookups take a long time. Please be patient._'
                             ).style('color: DarkOrange; margin-bottom: 0')]),
        ], [
            put_processbar('bar', init = 1/steps).style('margin-top: 11px'),
            put_button('Stop', outline = True, color = 'danger',
                       onclick = lambda: stop()).style('text-align: right'),
        ]], cell_widths = '85% 15%').style(PROGRESS_BOX)
        # The staff want to see location names, so we need to get the mapping.
        _running = True
        for count, id in enumerate(identifiers, start = 2):
            if _interrupted:
                break
            try:
                # Figure out what kind of identifier we were given.
                id_kind = folio.id_kind(id)
                if id_kind is IdKind.UNKNOWN:
                    tell_failure(f'Unrecognized identifier: **{id}**.')
                    continue
                if reuse_results:
                    records = _last_results.get(id)
                else:
                    records = folio.related_records(id, id_kind, kind_wanted,
                                                    pin.inventory_api, pin.open_loans)
                    _last_results[id] = records
                if not records or len(records) == 0:
                    tell_failure(f'No {kind_wanted} record(s) found for {id_kind} **{id}**.')
                    continue

                # Report the results & how we got them.
                source = 'storage'
                if pin.inventory_api and kind_wanted in ['item', 'instance']:
                    source = 'inventory'
                this = pluralized(kind_wanted + f' {source} record', records, True)
                how = f'by searching for {id_kind} **{id}**.'
                tell_success(f'Found {this} {how}')
                show_index = (len(records) > 1)
                for index, record in enumerate(records, start = 1):
                    print_record(record, id, index, show_index, pin.show_raw == 'json')
                total_found += len(records)
            except Interrupted as ex:
                log('stopping due to interruption')
                _interrupted = True
            except Exception as ex:
                import traceback
                log('Exception info: ' + str(ex) + '\n' + traceback.format_exc())
                tell_failure(f'Error: ' + str(ex))
                stop_processbar()
                return
            finally:
                if not _interrupted:
                    set_processbar('bar', count/steps)
        stop_processbar()
        clear_scope('current_activity')
        if _interrupted:
            tell_warning('**Stopped**.')
        else:
            summary = (f'Found {total_found} {kind_wanted} records by looking up '
                       + pluralized('unique identifier', identifiers, True)
                       + '.')
            put_grid([[
                put_markdown(summary).style('margin-top: 6px'),
                put_button('Export', outline = True,
                           onclick = lambda: do_export(_last_results, kind_wanted),
                           ).style('text-align: right')
            ]]).style('margin: 1.5em 17px auto 17px')
        _running = False


def field(record, field_name, subfield_name = None, list_joiner = ', '):
    if field_name not in record.data:
        return ''
    if subfield_name:
        if subfield_name not in record.data[field_name]:
            return ''
        value = record.data[field_name][subfield_name]
    else:
        value = record.data[field_name]
    if isinstance(value, list) and list_joiner:
        return list_joiner.join(str(x) for x in value)
    else:
        return str(value)


def location(record, field_name):
    global _location_map
    if field_name not in record.data:
        return ''
    location_data = record.data[field_name]
    if isinstance(location_data, dict):
        if 'name' in location_data:
            return f'{location_data["name"]}  ({location_data["id"]})'
        else:
            return location_data["id"]
    elif location_data and location_data in _location_map:
        return f'{_location_map[location_data]}  ({location_data})'
    return '(unknown location)'


def notes(record, field_name):
    if field_name not in record.data:
        return ''
    notes = record.data[field_name]
    if isinstance(notes, str):
        return notes
    elif isinstance(notes, list):
        if len(notes) == 0:
            return ''
        elif isinstance(notes[0], dict):
            return '\n'.join(n['note'] for n in notes)
        else:
            return '\n'.join(str(note) for note in notes)
    else:
        return notes


def print_record(record, identifier, index, show_index, show_raw):
    log(f'printing {record.kind} record {record.id}')
    if show_index:
        put_markdown(f'{record.kind.title()} record #{index}:')

    if show_raw:
        put_code(pformat(record.data, indent = 2))
    elif record.kind is RecordKind.ITEM:
        # Caution: left-hand values contain nonbreaking spaces (invisible here).
        if 'title' in record.data:
            # Inventory record version.
            put_table([
                ['Title'                     , field(record, 'title')],
                ['Barcode'                   , field(record, 'barcode')],
                ['Call number'               , field(record, 'callNumber')],
                [f'{record.kind.title()} id' , field(record, 'id')],
                ['Effective location'        , location(record, 'effectiveLocation')],
                ['Permanent location'        , location(record, 'permanentLocation')],
                ['Status'                    , field(record, 'status', 'name')],
                ['Tags'                      , field(record, 'tags', 'tagsList')],
                ['Notes'                     , notes(record, 'notes')],
                ['HRID'                      , field(record, 'hrid')],
                ['Created'                   , field(record, 'metadata', 'createdDate')],
                ['Updated'                   , field(record, 'metadata', 'updatedDate')],
            ]).style('font-size: 90%; margin: auto 17px 1.5em 17px')
        else:
            # Storage record version.
            put_table([
                ['Barcode'                   , field(record, 'barcode')],
                ['Call number'               , field(record, 'itemLevelCallNumber')],
                [f'{record.kind.title()} id' , field(record, 'id')],
                ['Effective location'        , location(record, 'effectiveLocationId')],
                ['Permanent location'        , location(record, 'permanentLocationId')],
                ['Tags'                      , field(record, 'tags', 'tagsList')],
                ['Notes'                     , notes(record, 'notes')],
                ['HRID'                      , field(record, 'hrid')],
                ['Created'                   , field(record, 'metadata', 'createdDate')],
                ['Updated'                   , field(record, 'metadata', 'updatedDate')],
            ]).style('font-size: 90%; margin: auto 17px 1.5em 17px')
    elif record.kind is RecordKind.INSTANCE:
        # Caution: left-hand values contain nonbreaking spaces (invisible here).
        if field(record, 'classifications'):
            call_number = record.data['classifications'][0]['classificationNumber']
        else:
            call_number = ''
        if 'tags' in record.data:
            put_table([
                ['Title'                     , field(record, 'title')],
                ['Call number'               , call_number],
                [f'{record.kind.title()} id' , field(record, 'id')],
                ['Tags'                      , field(record, 'tags', 'tagsList')],
                ['Notes'                     , notes(record, 'notes')],
                ['HRID'                      , field(record, 'hrid')],
                ['Created'                   , field(record, 'metadata', 'createdDate')],
                ['Updated'                   , field(record, 'metadata', 'updatedDate')],
            ]).style('font-size: 90%; margin: auto 17px 1.5em 17px')
        else:
            put_table([
                ['Title'                     , field(record, 'title')],
                ['Call number'               , call_number],
                [f'{record.kind.title()} id' , field(record, 'id')],
                ['HRID'                      , field(record, 'hrid')],
                ['Notes'                     , notes(record, 'notes')],
                ['Created'                   , field(record, 'metadata', 'createdDate')],
                ['Updated'                   , field(record, 'metadata', 'updatedDate')],
            ]).style('font-size: 90%; margin: auto 17px 1.5em 17px')
    elif record.kind is RecordKind.HOLDINGS:
        # Caution: left-hand values contain nonbreaking spaces (invisible here).
        if 'effectiveLocationId' in record.data:
            put_table([
                [f'{record.kind.title()} id' , field(record, 'id')],
                ['HRID'                      , field(record, 'hrid')],
                ['Holdings type id'          , field(record, 'holdingsTypeId')],
                ['Instance id'               , field(record, 'instanceId')],
                ['Effective location'        , location(record, 'effectiveLocationId')],
                ['Permanent location'        , location(record, 'permanentLocationId')],
                ['Created'                   , field(record, 'metadata', 'createdDate')],
                ['Updated'                   , field(record, 'metadata', 'updatedDate')],
            ]).style('font-size: 90%; margin: auto 17px 1.5em 17px')
        else:
            put_table([
                [f'{record.kind.title()} id' , field(record, 'id')],
                ['HRID'                      , field(record, 'hrid')],
                ['Holdings type id'          , field(record, 'holdingsTypeId')],
                ['Instance id'               , field(record, 'instanceId')],
                ['Effective location'        , ''],
                ['Permanent location'        , location(record, 'permanentLocationId')],
                ['Created'                   , field(record, 'metadata', 'createdDate')],
                ['Updated'                   , field(record, 'metadata', 'updatedDate')],
            ]).style('font-size: 90%; margin: auto 17px 1.5em 17px')
    elif record.kind is RecordKind.USER:
        # Caution: left-hand values contain nonbreaking spaces (invisible here).
        put_table([
            ['Username'                  , field(record, 'username')],
            ['Barcode'                   , field(record, 'barcode')],
            [f'{record.kind.title()} id' , field(record, 'id')],
            ['Patron group'              , field(record, 'patronGroup')],
            ['Created'                   , field(record, 'metadata', 'createdDate')],
            ['Updated'                   , field(record, 'metadata', 'updatedDate')],
        ]).style('font-size: 90%; margin: auto 17px 1.5em 17px')
    elif record.kind is RecordKind.LOAN:
        if 'userId' in record.data:
            put_table([
                [f'{record.kind.title()} id' , field(record, 'id')],
                ['Status',                     field(record, 'status', 'name')],
                ['User id'                   , field(record, 'userId')],
                ['Item id'                   , field(record, 'itemId')],
                ['Loan date'                 , field(record, 'loanDate')],
                ['Due date'                  , field(record, 'dueDate')],
                ['Created'                   , field(record, 'metadata', 'createdDate')],
                ['Updated'                   , field(record, 'metadata', 'updatedDate')],
            ]).style('font-size: 90%; margin: auto 17px 1.5em 17px')
        else:
            put_table([
                [f'{record.kind.title()} id' , field(record, 'id')],
                ['Status',                     field(record, 'status', 'name')],
                ['User id'                   , ''],
                ['Item id'                   , field(record, 'itemId')],
                ['Loan date'                 , field(record, 'loanDate')],
                ['Due date'                  , field(record, 'dueDate')],
                ['Created'                   , field(record, 'metadata', 'createdDate')],
                ['Updated'                   , field(record, 'metadata', 'updatedDate')],
            ]).style('font-size: 90%; margin: auto 17px 1.5em 17px')


def user_wants_reuse():
    event = threading.Event()
    answer = False

    def clk(val):
        nonlocal answer
        answer = val
        event.set()

    pins = [
        put_text('The list of identifiers and the kind of record to retrieve'
                 + ' are unchanged from the previous lookup. Should the results'
                 + ' be reused, or should the identifiers be looked up again?'),
        put_html('<br>'),
        put_buttons([
            {'label': 'Reuse the results', 'value': True},
            {'label': 'Search again', 'value': False, 'color': 'secondary'},
        ], onclick = clk).style('float: left')
    ]
    popup(title = 'Should results be reused?', content = pins, closable = False)

    event.wait()
    close_popup()
    wait(0.5)                           # Give time for popup to go away.

    return answer


def do_export(results, record_kind):
    log(f'exporting {record_kind} {pluralized("record", results, True)}')
    # Results is a dictionary; each value is a list of records. Unwind it.
    all_records = [item for value in results.values() for item in value]
    export_records(all_records, record_kind)



'''
change_tab.py: implementation of the "Change records" tab

Copyright
---------

Copyright (c) 2021-2022 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

from   collections import namedtuple, defaultdict
from   commonpy.data_utils import unique, pluralized, flattened
from   commonpy.exceptions import Interrupted
from   commonpy.file_utils import exists, readable
from   commonpy.interrupt import wait, reset_interrupts, interrupt, interrupted
from   decouple import config
from   pywebio.input import input, select, checkbox, radio
from   pywebio.input import NUMBER, TEXT, input_update, input_group
from   pywebio.output import put_text, put_markdown, put_row, put_html
from   pywebio.output import toast, popup, close_popup, put_buttons, put_button
from   pywebio.output import use_scope, set_scope, clear, remove, put_warning
from   pywebio.output import put_info, put_table, put_grid, span, put_link
from   pywebio.output import put_tabs, put_image, put_scrollable, put_code
from   pywebio.output import put_processbar, set_processbar, put_loading
from   pywebio.output import put_column, put_scope, clear_scope
from   pywebio.pin import pin, pin_wait_change, put_input, put_actions
from   pywebio.pin import put_textarea, put_radio, put_checkbox, put_select
from   pywebio.session import run_js, eval_js
from   sidetrack import set_debug, log
import sys
import threading

from   foliage.base_tab import FoliageTab
from   foliage.exceptions import *
from   foliage.export import export_data
from   foliage.folio import Folio, RecordKind, IdKind, TypeKind, Record
from   foliage.folio import unique_identifiers, back_up_record
from   foliage.ui import confirm, notify, user_file
from   foliage.ui import PROGRESS_BOX, PROGRESS_TEXT
from   foliage.ui import tell_success, tell_warning, tell_failure, stop_processbar
from   foliage.ui import note_info, note_warn, note_error, tell_success, tell_failure


# Tab definition class.
# .............................................................................

class ChangeTab(FoliageTab):
    def contents(self):
        return {'title': 'Change records', 'content': tab_contents()}

    def pin_watchers(self):
        return {'chg_op': lambda value: update_tab(value)}


# Tab creation function.
# .............................................................................

def tab_contents():
    log(f'generating change tab contents')
    # FIXME what causes these diffs on windows?
    textarea_rows = 14 if sys.platform.startswith('win') else 13
    margin_adjust = 'margin-top: -1.1em' if sys.platform.startswith('win') else ''
    return [
        put_grid([[
            put_markdown('Input item and/or holdings identifiers'
                         + ' (i.e., barcodes, id\'s, or hrid\'s). All'
                         + ' records will be changed the same way. Changing a'
                         + ' holdings record will also change all its items.'),
            put_button('Upload', outline = True,
                       onclick = lambda: load_file()).style('text-align: right'),
        ]], cell_widths = 'auto 100px'),
        put_grid([[
            put_grid([
                [put_textarea('textbox_ids', rows = textarea_rows)],
            ]),
            put_grid([
                [put_text('Select the field to be changed:').style('margin-top: -0.5em')],
                [put_row([
                    put_button('Select', onclick = lambda: select_field_name()
                               ).style('text-align: left'),
                    put_textarea('chg_field', rows = 1, readonly = True),
                ], size = '95px auto').style('text-align: right')],
                [put_text('Select the action to perform:')],
                [put_radio('chg_op', inline = True,
                           options = [ ('Add value', 'add', True),
                                       ('Change value', 'change'),
                                       ('Delete value', 'delete')]
                           ).style(f'margin-bottom: 0.3em; {margin_adjust}')],
                [put_text('Current field value (records must match this):').style('opacity: 0.3')],
                [put_row([
                    put_button('Select', onclick = lambda: select_field_value('old')),
                    put_textarea('old_value', rows = 1, readonly = True),
                ], size = '95px auto').style('z-index: 8; opacity: 0.3')],
                [put_text('New value (field will be set to this):')],
                [put_row([
                    put_button('Select', onclick = lambda: select_field_value('new')),
                    put_textarea('new_value', rows = 1, readonly = True),
                ], size = '95px auto').style('z-index: 9')],
                ]).style('margin-left: 12px'),
        ]], cell_widths = '50% 50%').style('margin-top: 1em'),
        put_row([
            put_button('Change records', color = 'danger',
                       onclick = lambda: do_change()),
            put_button('Clear', outline = True,
                       onclick = lambda: clear_tab()).style('text-align: right'),
        ])
    ]


# Implementation of tab functionality.
# .............................................................................

# This next bit is an egregious and unobvious hack.  Here's the deal.  When
# the user selects different options using the radio buttons for the
# operation type (add value, change value, delete value), we want different
# buttons and fields to look visually "unavailable" as a clue to the user
# that they don't have to fill in those values.  However, PyWebIO doesn't
# provide a way to control the properties of the elements dynamically: there's
# PyWebIO API for changing a CSS attribute at run-time.
#
# One *can* do it using JavaScript using well-known methods, and PyWebIO
# *does* provide a function (eval_js) to execute JavaScript at run-time in
# the web page.  But, here we face a new challenge: how do do you refer to
# the things on the page whose CSS attributes you want to change?
#
# For some of the elements, it's possible to target them by searching for the
# element content.  That's the case for the text fields, where we can use a
# jQuery selector such as
#    $("p.contains('Current field value')")
# to get at the element, and from there, to change the CSS.  This is used in
# the code below for elements for which it's possible to do that.  But you
# can't do that for buttons -- you need to find another way to refer to them.
# Frustratingly, PyWebIO doesn't provide a way to put id attributes on
# elements; if you could do that, it would make it easy to target exactly the
# elements you need to change.  You also can't target CSS classes, because
# that would end up catching other elements with the same class on the page.
#
# So to get the other elements (the ones that can't be found using the jQuery
# "contains" operator mentioned above), I ended up using an insane hack:
#
#  1. Add distinguishing features to specific elements using one of the few
#     CSS/HTML controls that PyWebIO does provide, namely the style()
#     function, to add a property that we can uniquely find in the DOM at run
#     time.  I'm using the z-index, setting it specific numbers (8 and 9) in
#     tab_contents() above.  The z-index is not used for anything else on
#     this page so it's irrelevant as far as layout is concerned.
#
#  2. Invoke some JavaScript code in the web page that uses jQuery to look
#     for the elements that have the specific z-index values.  That's how the
#     specific elements are found.  Then it's easy to change the value of
#     desired CSS properties on those elements.
#
# Why use the numbers 8 and 9?  In case a future change ends up using a z-index
# value for something.  I hypothesized that a future developer who used z-index
# for a real purpose would use a value like 1, 2, or maybe 100, 1000, etc.
# The values 8 and 9 seemed off-beat enough that they wouldn't clash with
# something in the future, even if a future developer doesn't read this comment
# explaining what's going on.

def update_tab(value):
    log(f'updating form in response to radio box selection: "{value}"')
    if value == 'add':
        eval_js('''$("p:contains('Current field value')").css("opacity", "0.3");''')
        eval_js('''$("div").filter((i, n) => $(n).css("z-index") == 8).css("opacity", "0.3");''')
        eval_js('''$("p:contains('New field value')").css("opacity", "1");''')
        eval_js('''$("div").filter((i, n) => $(n).css("z-index") == 9).css("opacity", "1");''')
    elif value == 'delete':
        eval_js('''$("p:contains('Current field value')").css("opacity", "1");''')
        eval_js('''$("p:contains('New field value')").css("opacity", "0.3");''')
        eval_js('''$("div").filter((i, n) => $(n).css("z-index") == 8).css("opacity", "1");''')
        eval_js('''$("div").filter((i, n) => $(n).css("z-index") == 9).css("opacity", "0.3");''')
    else:
        eval_js('''$("p:contains('Current field value')").css("opacity", "1");''')
        eval_js('''$("div").filter((i, n) => $(n).css("z-index") == 8).css("opacity", "1");''')
        eval_js('''$("p:contains('New field value')").css("opacity", "1");''')
        eval_js('''$("div").filter((i, n) => $(n).css("z-index") == 9).css("opacity", "1");''')


def clear_tab():
    log(f'clearing tab')
    clear('output')
    pin.textbox_ids = ''
    pin.chg_op = 'add'
    pin.chg_field = ''
    pin.old_value = ''
    pin.new_value = ''
    update_tab('add')


def select_field_name():
    # Clear any previous value.
    pin.new_value = ''
    if (answer := selected('Select field to change', known_fields)):
        # Show the selected value.
        pin.chg_field = answer
        log(f'user selected field {answer}')


def select_field_value(old_new):
    # No way to prevent clicks when the op is not valid, so just ignore them.
    # Setting an old field value is only valid for change and delete.
    # Setting a new field value is only valid for add and change.
    if ((old_new == 'old' and pin.chg_op == 'add')
        or (old_new == 'new' and pin.chg_op == 'delete')):
        return

    if not pin.chg_field:
        notify('Please first select the field to be changed.')
        return

    fname = pin.chg_field.lower()
    log(f'getting list of values for {fname}')
    type_list = Folio().types(known_fields[pin.chg_field].type)
    if not type_list:
        note_error(f'Could not retrieve the list of possible {fname} values')
        return
    value_list = sorted(item.data['name'] for item in type_list)
    if (val := selected(f'Select the {old_new} value for {fname}', value_list)):
        field = old_new + '_value'
        setattr(pin, field, val)
        log(f'user selected {old_new} field value {val}')


def selected(title, values):
    log(f'showing list selection popup')
    event = threading.Event()
    clicked_ok = False

    def clk(val):
        nonlocal clicked_ok
        clicked_ok = val
        event.set()

    pins = [
        put_select('list_selection', options = values),
        put_buttons([
            {'label': 'Submit', 'value': True},
            {'label': 'Cancel', 'value': False, 'color': 'secondary'},
        ], onclick = clk).style('float: right')
    ]
    popup(title = title, content = pins, closable = False)
    event.wait()
    close_popup()
    wait(0.5)                           # Give time for popup to go away.

    log(f'user {"made a selection" if clicked_ok else "cancelled"}')
    return pin.list_selection if clicked_ok else None


def all_selections_made():
    return (pin.chg_field
            and ((pin.chg_op == 'add' and pin.new_value)
                 or (pin.chg_op == 'delete' and pin.old_value)
                 or (pin.chg_op == 'change' and pin.new_value and pin.old_value)))


def load_file():
    log(f'user requesting file upload')
    if (contents := user_file('Upload a file containing identifiers')):
        pin.textbox_ids = contents


def stop():
    log(f'stopping')
    interrupt()
    stop_processbar()


_results = []

def clear_results():
    global _results
    _results = []


def record_result(record_or_id, success, notes):
    global _results
    id = record_or_id if isinstance(record_or_id, str) else record_or_id.id
    rec = record_or_id if isinstance(record_or_id, Record) else None
    _results.append({'id': id, 'success': success, 'notes': notes, 'record': rec})


def succeeded(record_or_id, msg, context = ''):
    comment = (' (' + context + ')') if context else ''
    record_result(record_or_id, True, msg + comment)
    tell_success(f'Success for **{id}**{comment}: ' + msg + '.')


def failed(record_or_id, msg, context = ''):
    comment = (' (' + context + ')') if context else ''
    record_result(record_or_id, False, msg + comment)
    id = record_or_id if isinstance(record_or_id, str) else record_or_id.id
    tell_failure(f'Failure for **{id}**{comment}: ' + msg + '.')


def skipped(record_or_id, msg, context = ''):
    comment = (' (' + context + ')') if context else ''
    record_result(record_or_id, False, msg + comment)
    id = record_or_id if isinstance(record_or_id, str) else record_or_id.id
    tell_warning(f'Skipped **{id}**{comment}: ' + msg + '.')


_SUPPORTED_KINDS = [
    RecordKind.ITEM,
    RecordKind.HOLDINGS,
]

def do_change():
    log(f'do_change invoked')
    identifiers = unique_identifiers(pin.textbox_ids)
    if not identifiers:
        note_error('Please input at least one barcode or other type of id.')
        return
    if not all_selections_made():
        note_error('Missing selections – cannot proceed until form is filled out.')
        return
    if not confirm('Proceed with changes to records in FOLIO?', danger = True):
        log(f'user declined to proceed')
        return
    clear_results()
    reset_interrupts()
    steps = 2*len(identifiers)         # We need 2 passes => 2x number of items
    folio = Folio()
    with use_scope('output', clear = True):
        try:
            done = 0
            put_grid([[
                put_scope('current_activity', [
                    put_markdown(f'_Gathering records ..._').style(PROGRESS_TEXT)]),
            ], [
                put_processbar('bar', init = done/steps).style('margin-top: 11px'),
                put_button('Stop', outline = True, color = 'danger',
                           onclick = lambda: stop()).style('text-align: right')
            ]], cell_widths = '85% 15%').style(PROGRESS_BOX)

            # Start by gathering all records & their types.
            records = []
            for id in identifiers:
                record = folio.record(id)
                done += 1
                set_processbar('bar', done/steps)
                if not record:
                    failed(id, f'unrecognized identifier **{id}**')
                    continue
                if record.kind not in _SUPPORTED_KINDS:
                    skipped(id, f'changing {record.kind} records is not supported')
                    continue
                records.append(record)

            # 1st pass: apply changes to holdings records in the input (if any).
            with use_scope('current_activity', clear = True):
                put_markdown(f'_Changing records ..._').style(PROGRESS_TEXT)
            holdings_done = []
            for record in filter(lambda r: r.kind is RecordKind.HOLDINGS, records):
                done += 1
                set_processbar('bar', done/steps)
                if not change_holdings(record):
                    log(f'couldn\'t change and/or save holdings rec. – skipping items')
                    continue
                for item in folio.related_records(record.id, IdKind.HOLDINGS_ID,
                                                  RecordKind.ITEM):
                    log(f'changing item {item.id} after changing holdings {record.id}')
                    context = f'an item associated with holdings record {record.id}'
                    # We changed the holdings rec. => we can change item directly.
                    change_item(item, record, context = context)
                holdings_done.append(record.id)

            # 2nd pass: apply changes to item records in the input. Some may
            # have been changed in 1st pass if the user provided holdings recs.
            for item in filter(lambda r: r.kind is RecordKind.ITEM, records):
                if item.data['holdingsRecordId'] in holdings_done:
                    log(f'skipping {item.id}, assuming it was done in holdings pass')
                change_item(item)
                done += 1
                set_processbar('bar', done/steps)
            set_processbar('bar', 1)
        except Interrupted as ex:
            tell_warning('**Stopped**.')
            return
        except Exception as ex:
            import traceback
            log('Exception info: ' + str(ex) + '\n' + traceback.format_exc())
            tell_failure(f'Error: ' + str(ex))
            return
        finally:
            stop_processbar()
            clear_scope('current_activity')

        what = pluralized('record', identifiers, True)
        put_grid([[
            put_markdown(f'Finished changing {what}.').style('margin-top: 6px'),
            put_button('Export summary', outline = True,
                       onclick = lambda: do_export('foliage-changes.csv'),
                       ).style('text-align: right')
        ]]).style('margin: auto 17px auto 17px')


def change_holdings(record):
    if pin.chg_field == 'Permanent location' and pin.chg_op != 'change':
        # Holdings always have a perm loc., so can only change it.
        skipped(record.id, 'Cannot add or delete a permanent'
                + ' location field on holdings records')
        return False
    return change_record(record) and save_changes(record)


def change_item(item, given_hrec = None, context = ''):
    '''Change the item and also update its parent holdings.'''

    # Try to change the item but without saving it yet. If we fail, bail.
    if not change_record(item, context):
        return False
    else:
        save_changes(item)

    # If the change is to a temporary location field, we can make the change
    # without having to change a holdings record, and we're done.
    if pin.chg_field == 'Temporary location':
        return True

    # Not a temporary location change. Were we given the holdings record in the
    # input too? Then we should have done the change already.
    field_key = known_fields[pin.chg_field].key
    if given_hrec:
        if item.data[field_key] == given_hrec.data[field_key]:
            return True
        else:
            # This should not happen b/c the caller should have set its loc.
            failed(item.id, 'inconsistency in Foliage -- please report this.')
            return False

    # No holdings record has been changed so far. Here are the possible cases:
    # 1. Moving item from one holdings to another holdings on the instance.
    #    1a) Original holdings has other items => no deletion needed. Done.
    #    1b) Original holdings is empty => delete original holdings record.
    #
    # 2. Moving item to a location that has no holdings record on the instance
    #    2a) The instance has only one item => change the existing holdings rec.
    #    2b) The instance has other items => create new holdings record.
    location_id = item.data['permanentLocationId']
    folio = Folio()
    hrec = folio.record(item.data['holdingsRecordId'])
    if not hrec:
        # This should never happen, but we always want to check everything.
        failed(item.id, f'cannot update {field_key} of {item.id} because its holdings'
               + f' record {item.data["holdingsRecordId"]} could not be retrieved')
        return False

    # In practice, the only way we get to this point is if the operation is not
    # a deletion -- so either an add or a change.
    inst_id = hrec.data['instanceId']
    all_hrecs = folio.related_records(inst_id, IdKind.INSTANCE_ID, RecordKind.HOLDINGS)
    hrec_items = None
    log(f'parent instance {inst_id} has {len(all_hrecs)} holdings records')
    for h in all_hrecs:
        log(f'checking location of holdings record {h.id}')
        if item.data[field_key] == h.data[field_key]:
            # Found a holdings record that has the new location => case 1.
            log(f'updating {item.id}\'s holdings record to be {h.id}')
            item.data['holdingsRecordId'] = h.id
            if config('DEMO_MODE', cast = bool):
                log(f'demo mode – pretending to save {item.id}')
            else:
                try:
                    folio.update_record(item)
                except FolioOpFailed as ex:
                    failed(item.id, str(ex), context)
                    return False
            context = f'updating item\'s holdings record pointer'
            succeeded(item.id, f'changed holdings record to be {h.id}', context)
            break
    else:
        log('none of the holdings record have the new location')
        # We have case 2. Next check: does the instance have any other items?
        if len(all_hrecs) == 1:
            hrec_items = folio.related_records(hrec.id, IdKind.HOLDINGS_ID, RecordKind.ITEM)
            if len(hrec_items) == 1:
                context = f'holdings record for {item.id}'
                # Case 2a: the instance has only 1 item.
                if change_holdings(hrec):
                    succeeded(hrec.id, f'field _{field_key}_ changed', context)
                    return True
                else:
                    failed(item.id, 'failed to change holdings record')
                    return False

        # Case 2b: the instance has other items. We must create a new holdings
        # record. Do it by copying the existing one & modifying it.
        log(f'need to create new holdings record for moving {item.id}')
        import copy
        new_holdings = copy.deepcopy(hrec)
        # These next fields are assigned automatically the Folio server.
        del new_holdings.data['id']
        del new_holdings.data['hrid']
        del new_holdings.data['metadata']
        # Update fields for the move.
        new_holdings.data['permanentLocationId'] = location_id
        # Create the record.
        context = f'moving item to new holdings record for location {location_id}'
        if config('DEMO_MODE', cast = bool):
            log(f'demo mode – pretending to create new holdings record')
            new_id = '[holdings id]'
        else:
            try:
                new_id = folio.new_record(new_holdings)
            except FolioOpFailed as ex:
                failed(item.id, str(ex), context = context)
                return False
        succeeded(item.id, f'created holdings record {new_id}', context)

        log(f'changing location of {item.id} to new holdings record {new_id}')
        item.data['holdingsRecordId'] = new_id
        if config('DEMO_MODE', cast = bool):
            log(f'demo mode – pretending to save {item.id}')
        else:
            try:
                back_up_record(item)
                folio.update_record(item)
            except FolioOpFailed as ex:
                failed(item.id, str(ex), context)
                return False
        succeeded(item.id, f'attached item to holdings record {new_id}', context)

    # We've moved the item. Do we need to delete the holdings record it came
    # from? Check if there are any other items on it.
    if hrec_items is None:
        hrec_items = folio.related_records(hrec.id, IdKind.HOLDINGS_ID, RecordKind.ITEM)
    for other in filter(lambda record: record.id != item.id, hrec_items):
        if other.id != item.id:
            log('holdings record has other items, therefore not deleting it')
            return True

    # It's 1b (orig holdings rec is now empty). Need delete the holdings rec.
    id = hrec.id
    context = 'moved last or only item to another holdings record'
    if config('DEMO_MODE', cast = bool):
        log(f'demo mode – pretending to delete {id}')
    else:
        try:
            back_up_record(hrec)
            folio.delete_record(hrec)
        except FolioOpFailed as ex:
            failed(id, str(ex), context = context)
            return False
    succeeded(item.id, f'deleted empty holdings record {id}', context)
    return True


def change_record(record, context = ''):
    '''Adds, changes, or deletes a field value in the record.
    The change is determined by the current UI selections (via the "pins").
    Returns True if successful, False if couldn't make the change.
    Does not save the record; a save is assumed to be performed by the caller.
    '''
    folio = Folio()
    # Get the list of known values for this type (folio.py will have cached
    # it) and create a mapping of value names to value objects.
    value_type = known_fields[pin.chg_field].type
    values = {x.data['name']:x.data for x in folio.types(value_type)}

    field_key = known_fields[pin.chg_field].key
    if (current_value := record.data.get(field_key, None)):
        if pin.chg_op == 'add':
            skipped(record.id, f'item _{field_key}_ has an existing value', context)
            return False
        # We're doing change or delete. Existing value must match expected one.
        if current_value != values[pin.old_value]['id']:
            skipped(record.id, f'_{field_key}_ value is not "{pin.old_value}"', context)
            return False
        # We can proceed.
        back_up_record(record)
        if pin.chg_op == 'change':
            log(f'changing field {field_key} in item {record.id} to {pin.new_value}')
            record.data[field_key] = values[pin.new_value]['id']
        elif pin.chg_op == 'delete':
            # Some fields on some record kinds must always exist.
            if record.kind in known_fields[pin.chg_field].delete_ok:
                log(f'deleting field {field_key} from item {record.id}')
                del record.data[field_key]
            else:
                skipped(record.id, f'not allowed to delete field {field_key}'
                        + ' on {record.kind} records', context)
                return False
    elif pin.chg_op == 'add':
        log(f'adding {field_key} value "{pin.new_value}" to item {record.id}')
        back_up_record(record)
        record.data[field_key] = values[pin.new_value]['id']
    else:
        # It doesn't have a value, and we're not doing an add.
        skipped(record.id, f'item **{record.id}** has no field _{field_key}_', context)
        return False
    log(f'changes made to {record.kind} record {record.id}')
    return True


def save_changes(record, context = ''):
    if config('DEMO_MODE', cast = bool):
        log(f'demo mode – pretending to save {record.id}')
    else:
        try:
            folio = Folio()
            folio.update_record(record)
        except FolioOpFailed as ex:
            failed(record.id, str(ex), context)
            return False

    # Report the outcome to the user.
    field = 'field _' + known_fields[pin.chg_field].key + '_'
    text = {'add': 'added to', 'change': 'changed in', 'delete': 'deleted from'}
    action = text[pin.chg_op]
    succeeded(record.id, f'{field} {action} {record.kind} record', context)
    return True


def do_export(file_name):
    global _results
    values = []
    # WIP need to add more fields, but that requires more changes.
    for result in _results:
        entry = {'Record ID'          : result['id'],
                 'Operation success'  : result['success'],
                 'Notes'              : result['notes']}
        values.append(entry)
    export_data(values, file_name)


Field = namedtuple('Field', 'type key delete_ok')

known_fields = {
    'Temporary location': Field(type = TypeKind.LOCATION,
                                key = 'temporaryLocationId',
                                delete_ok = [RecordKind.ITEM, RecordKind.HOLDINGS]),
    'Permanent location': Field(type = TypeKind.LOCATION,
                                key = 'permanentLocationId',
                                delete_ok = [RecordKind.ITEM]),
}



'''
other_tab.py: implementation of the "Other" tab

Copyright
---------

Copyright (c) 2021-2022 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

from   commonpy.data_utils import unique, pluralized, flattened
from   commonpy.file_utils import exists, readable, open_file
from   commonpy.interrupt import wait
from   decouple import config
from   os.path import dirname
from   pywebio.input import input, select, checkbox, radio
from   pywebio.input import NUMBER, TEXT, input_update, input_group
from   pywebio.output import put_text, put_markdown, put_row, put_html
from   pywebio.output import toast, popup, close_popup, put_buttons, put_button, put_error
from   pywebio.output import use_scope, set_scope, clear, remove, put_warning
from   pywebio.output import put_success, put_info, put_table, put_grid, span
from   pywebio.output import put_tabs, put_image, put_scrollable, put_code, put_link
from   pywebio.output import put_processbar, set_processbar, put_loading
from   pywebio.output import put_column
from   pywebio.pin import pin, pin_wait_change, put_input, put_actions
from   pywebio.pin import put_textarea, put_radio, put_checkbox, put_select
from   sidetrack import set_debug, log
from   tempfile import NamedTemporaryFile
import threading
import webbrowser

from   foliage import __version__
from   foliage.base_tab import FoliageTab
from   foliage.credentials import credentials_from_user, current_credentials
from   foliage.credentials import use_credentials
from   foliage.folio import Folio, RecordKind, IdKind, TypeKind
from   foliage.ui import quit_app, reload_page, confirm, notify
from   foliage.ui import image_data, user_file, JS_CODE, CSS_CODE
from   foliage.ui import note_info, note_warn, note_error, tell_success, tell_failure


# Tab definition class.
# .............................................................................

class OtherTab(FoliageTab):
    def contents(self):
        return {'title': 'Other', 'content': tab_contents()}

    def pin_watchers(self):
        return {}


# Tab creation function.
# .............................................................................

def tab_contents():
    log(f'generating other tab contents')
    return [
        put_html(f'<i><b>Help for Foliage</b> is <a target="_blank"'
                 + ' href="https://caltechlibrary.github.io/foliage"'
                 + '>available online</a></i>.'
                     ).style('text-align: center; margin-bottom: 1em'),
        put_grid([[
            put_markdown('Foliage stores the FOLIO credentials you provide the'
                         + ' first time it runs, so that you don\'t have to'
                         + ' enter them again. Click this button to update the'
                         + ' stored credentials.'),
            put_button('Edit credentials', onclick = lambda: edit_credentials(),
                       ).style('margin-left: 20px; text-align: left'),
        ], [
            put_markdown('Before performing destructive operations, Foliage'
                         + ' saves copies of the records as they exist before'
                         + ' modification. Click this button to open the folder'
                         + ' containing the files. (Note: a given record may'
                         + ' have multiple backups with different time stamps.)'),
            put_button('Show backups', onclick = lambda: show_backup_dir(),
                       ).style('margin-left: 20px; margin-top: 0.8em'),
        ], [
            put_markdown('The debug log file contains a detailed trace of'
                         + ' every action that Foliage takes. This can be'
                         + ' useful when trying to resolve bugs and other'
                         + ' problems.'),
            put_button('Show log file', onclick = lambda: show_log_file(),
                       ).style('margin-left: 20px; text-align: left'),
        ]], cell_widths = 'auto 170px', cell_heights = '29% 42% 29%'),
        put_markdown(f'_You are running Foliage version {__version__}_.'
                     ).style('margin-top: 1em; text-align: center')
    ]


# Miscellaneous helper functions.
# .............................................................................

def edit_credentials():
    log('user invoked Edit credentials')
    current = current_credentials()
    creds = credentials_from_user(warn_empty = False, initial_creds = current)
    if creds and creds != current:
        log('user has provided updated credentials')
        use_credentials(creds)
    else:
        note_warn('Credentials unchanged')


def show_backup_dir():
    log(f'user invoked Show backup dir')
    webbrowser.open_new("file://" + config('BACKUP_DIR'))


def show_log_file():
    log_file = config('LOG_FILE')
    if log_file == '-':
        note_warn('No log file -- log output is being directed to the terminal.')
        return

    def close_window():
        event.set()

    def open_log_folder():
        open_file(dirname(log_file))

    event = threading.Event()
    log_contents = ''
    log('reading log file ' + log_file)
    with open(log_file, 'r') as f:
        log_contents = f.read()
    pins  = [
        put_scrollable(put_code(log_contents).style('font-size: smaller'),
                       height = 400, horizontal_scroll = True),
        put_grid([[
            put_button('Open log folder', outline = True,
                       onclick = lambda: open_log_folder()),
            put_button('Close', onclick = close_window).style('text-align: right')
        ]])
    ]
    popup(title = log_file, content = pins, implicit_close = True, size = 'large')

    event.wait()
    close_popup()



'''
ui.py: user interface utilities for Foliage

This contains miscellaneous user interface code:

* some CSS styling for Foliage
* some JavaScript code that gets added to the Foliage web page to work around
    some undesirable behaviors in PyWebIO, sometimes by calling jQuery
    functions to do something in the DOM
* some functions to encapsulate operations in PyWebIO.
* some user interface utility functions, including PyQt-based message functions
    for use before the main Foliage PyWebIO window is available

Comments in the rest of this file try to explain what is going on in some
places.

Copyright
---------

Copyright (c) 2021-2022 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

from   commonpy.data_utils import flattened
from   commonpy.interrupt import wait
from   commonpy.string_utils import antiformat
from   contextlib import contextmanager
from   decouple import config
import os
from   os.path import exists, dirname, join, basename, abspath
from   PyQt5.QtWidgets import QApplication, QMessageBox
import pywebio
from   pywebio.input import input, file_upload
from   pywebio.output import put_text, put_markdown, put_row, put_html
from   pywebio.output import toast, popup, close_popup, put_buttons
from   pywebio.output import put_success, put_warning, put_error
from   pywebio.output import use_scope, set_scope, clear, remove
from   pywebio.output import PopupSize, put_loading
from   pywebio.pin import pin, pin_wait_change, put_input, put_actions
from   pywebio.session import run_js, eval_js
from   rich.panel import Panel
from   rich.style import Style
import sys
import threading

if __debug__:
    from sidetrack import set_debug, log


# Exported constants.
# .............................................................................

UPLOAD_CANCEL_MARKER = "_fake_foliage_fake_.txt"

JS_CODE = '''
function close_window() { window.close() }
function reload_page() { location.reload() }


/* Make the escape key work in file upload dialogs.  The natural action (IMHO)
   for ESC is to cancel the dialog, but there's no "cancel" functionality in
   PyWebIO's file_upload() dialog.  In fact, if you click the "reset" button
   then click "submit", you still get a file!  Very undesirable behavior.
   There's no good way to fix it short of rewriting file.ts in the PyWebIO
   code, so the following is an egregious hack: muck with the variable that
   the code uses to store the file from the file input dialog, specifically by
   setting the variable to a known fake name when ESC is pressed.

   The solution to setting the .files property (which is a read-only FileList
   object) came from a 2019-06-04 posting by "superluminary" to Stack Overflow
   at https://stackoverflow.com/a/56447852/743730
*/
$(document).keyup(function(e) {
    /* Ignore this key press if it's not the escape key. */
    if (e.keyCode != 27) return;

    /* Case of PyWebIO modal dialogs with a cancel button. */
    if ($('.modal-content button.btn-secondary').length) {
        $('.modal-content button.btn-secondary').click();
    }

    /* Case of PyWebIO file_upload() dialog, which lacks a cancel button. */
    if ($('.ws-form-submit-btns button[type="reset"]').length) {
        // Create a fake FileList object and reset the .files property.
        let tmp_list = new DataTransfer();
        let fake = new File(["content"], "%s");
        tmp_list.items.add(fake);
        let myFileList = tmp_list.files;
        $('#input-cards .custom-file')[0].firstElementChild.files = myFileList;
        console.log($('#input-cards .custom-file')[0].firstElementChild.files);

        // Pretend the user clicked the "reset" button.
        $('.ws-form-submit-btns button[type="reset"]').click();

        // Give it a short time for JavaScript actions to work, and submit.
        setTimeout(() => { $('.custom-file input').submit() }, 200);
    }
});
''' % UPLOAD_CANCEL_MARKER

CSS_CODE = '''
html {
    position: relative;
}

body {
    padding-bottom: 66px;
}

h1 {
    font-size: 26pt;
}

.text-muted b {
    color: #555;
}

pre {
    background-color: #f8f8f8;
    padding: 5px;
    border: 1px solid #eee;
    border-radius: .25rem;
}

.pywebio {
    min-height: calc(100vh - 70px);
    padding-top: 10px;
    padding-bottom: 1px; /* if set 0, safari has min-height issue */
}

.pywebio_cancel_btn {
    float: right;
}

#output-container {
    margin-bottom: 40px;
}

footer {
    position: absolute;
    width: 100%;
    bottom: -1px;
    height: 58px !important;
    background-color: white !important;
}

.markdown-body table {
    display: inline-table;
}

.markdown-body table td {
    padding: 3px;
    padding-right: 6px;
}

table tbody tr {
    border-bottom: 1px dotted #e0e0e0;
}

.alert p {
    margin-bottom: 0
}

button {
    margin-bottom: 0 !important;
    filter: drop-shadow(1px 1px 2px #eee);
}

.btn {
    margin-bottom: 1px !important;
    min-width: 85px;
}

.btn-link {
    padding: 0
}

/* Weird 1px vertical misalignment. Don't know why I have to do this. */

.btn-danger, .btn-primary, .btn-secondary {
    margin-bottom: 1pt !important;
}

/* Special case for danger buttons, to adjust due to effects of red color. */
.btn-danger {
    filter: drop-shadow(1px 1px 2px #dadada);
}

.webio-tabs-content {
    padding-bottom: 0 !important;
}

#input-container {
    border-radius: .25rem;
    box-shadow: 10px 10px 20px #aaa;
    position: absolute;
    padding-left: 0;
    padding-right: 0;
    top: 190px;
    width: 500px;
    left: calc(50% - 250px);
}

#input-cards.container {
    padding-left: 0 !important;
    padding-right: 0 !important;
    border-radius: .25rem;
}

textarea.form-control[readonly] {
    background: repeating-linear-gradient(-45deg, #fff, #f6f6f6 8px);
}

.spinner-border {
    position: absolute;
    left: calc(50% - 1em);
    top: 7em;
}

.modal-lg {
    max-width: 90%;
}

.disabled-button {
    pointer-events: none;
    color: #ccc;
    border-color: #ccc;
}
'''

PROGRESS_BOX = '''
padding: 17px;
margin-bottom: 17px;
border: 1px solid #ededed;
border-radius: 0.25em
'''

PROGRESS_TEXT = '''
margin: 0;
font-size: 90%;
'''

EXCEL_MIME_TYPES = [
    'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
    'application/vnd.ms-excel'
]


# Exported functions
# .............................................................................

def inside_pyinstaller_app():
    '''Return True if we are running as an app created using PyInstaller.'''
    # This function is for the sake of making code more readable, because
    # the purpose of testing the following condition is not at all obvious.
    return getattr(sys, 'frozen', False)


def close_splash_screen():
    '''Close the PyInstaller-based splash screen shown during startup.'''
    if not inside_pyinstaller_app():
        return
    if sys.platform.startswith('darwin'):
        # PyInstaller does not currently support splash screens on macOS.
        return
    log('closing splash screen')
    try:
        # pyi_splash only exists inside the PyInstaller-produced executable.
        import pyi_splash
        pyi_splash.close()
    except Exception as ex:
        # Only log an error if running the PyInstaller-produced app.
        if inside_pyinstaller_app():
            log('exception trying to close splash screen: ' + str(ex))


def confirm(question, danger = False):
    log(f'asking user to confirm: {antiformat(question)}')

    event = threading.Event()
    clicked_ok = False

    def clk(val):
        nonlocal clicked_ok
        clicked_ok = val
        event.set()

    ok_color = 'danger' if danger else 'primary'
    pins = [
        put_buttons([
            {'label': 'Cancel', 'value': False, 'color': 'secondary'},
            {'label': 'OK',     'value': True, 'color': ok_color},
        ], onclick = clk).style('float: right')
    ]
    popup(title = '⚠️ ' + question, content = pins, closable = False)
    event.wait()
    close_popup()
    wait(0.25)                           # Give time for popup to go away.

    log(f'user clicked {"OK" if clicked_ok else "Cancel"}')
    return clicked_ok


def notify(msg):
    log(f'notifying user with message "{antiformat(msg)}"')

    event = threading.Event()

    def clk(val):
        event.set()

    pins = [
        put_buttons([{'label': 'OK', 'value': True}],
                    onclick = clk).style('float: right')
    ]
    popup(title = '✋ ' + msg, content = pins, closable = True)
    event.wait()
    close_popup()
    wait(0.25)                           # Give time for popup to go away.
    log(f'notification popup closed explicitly')


def stop_processbar():
    '''Stop the animation of the PyWebIO process bar.'''
    # PyWebIO uses Bootstrap animation for the progress bar.  Nice ... except
    # it doesn't provide a way to *stop* the animation effect!  Without that,
    # the progress bar maintains its movement animation even if the bar reaches
    # 100%, or you interrupt the operation.  The following code uses jQuery to
    # find the processbar via its id (which is set by PyWebIO) and remove the
    # Bootsrap class that controls the animation state.
    eval_js('''$("#webio-processbar-bar").removeClass("progress-bar-animated");''')
    # In all the Foliage pages with process bars, there's also a stop button.
    # When we stop the process bar, mute the button too.
    eval_js('''$("button:contains('Stop')").addClass("disabled-button");''')


def quit_app(ask_confirm = True):
    log(f'quitting (ask = {ask_confirm})')
    if ask_confirm:
        wait(0.25)
    if not ask_confirm or confirm('Exit Foliage?', danger = True):
        log(f'running JS function close_window()')
        run_js('close_window()')
        wait(0.5)
        raise SystemExit('User quit application')


def reload_page():
    '''Reload the Foliage application page.'''
    log(f'running JS function to reload the page')
    run_js('reload_page()')


def image_data(file_name):
    '''Return the data from the given image file.'''
    here = dirname(__file__)
    image_file = join(here, 'data', file_name)
    if exists(image_file):
        log(f'reading image file {antiformat(image_file)}')
        with open(image_file, 'rb') as f:
            return f.read()
    log(f'could not find image in {antiformat(image_file)}')
    return b''


def user_file(instructions):
    '''Ask the user to upload a file and return the contents as text.
    Currently supports plain text, CSV, and MS Office .xslx files.
    '''
    result = file_upload(instructions,
                         help_text = 'The file can be in any of the following'
                         + ' formats: .txt (plain text), .csv (comma-separated'
                         + '  values), .xlsx (Excel).')
    if result and result['filename'] != UPLOAD_CANCEL_MARKER:
        if result['mime_type'] in ['text/plain', 'text/csv']:
            return result['content'].decode()
        elif result['mime_type'] in EXCEL_MIME_TYPES:
            # It's an excel .xsl or .xslx file.
            import io
            from openpyxl import load_workbook
            content = io.BytesIO(result['content'])
            wb = load_workbook(content)
            ws = wb.active
            # The rows will be tuples. Flatten everything out as text.
            try:
                return '\n'.join(flattened(ws.values))
            except Exception as ex:
                log('failed to get flattened worksheet: ' + str(ex))
                notify('Unable to extract values from this spreadsheet.'
                       + ' This is probably an error in Foliage. Please'
                       + ' report it to the developers.')
        else:
            notify('This type of file is currently unsupported.'
                   + f' (MIME type {result["mime_type"]}.) Please contact the'
                   + 'developers to request support for this type.')
            return None
    return None


# The remainder of this file implements functions for reporting info, warning,
# success and failure, using 2 separate approaches. Summary of the scheme:
#
# 1) the tell_* functions print a message in the output area
# 2) the note_* functions print a "toast" message temporarily across the top
#
# Warning & error note_* functions print to the console if popup == False.
# The note_info(...) function just doesn't print anything if popup == False.
#
# Since note_* functions may be called before the PyWebIO GUI system has
# started, we need alternative approaches for different cases.
#
#                 Has the PyWebIO GUI started?
#                          /           \
#                       yes            no
#                       /               \
#    Use PyWebIO functions     Are we in the PyInstaller-built app?
#                                        /        \
#                                      yes        no
#                                      /           \
#                                 Use PyQt      Print to command line
#
# This uses PyQt message widgets as a platform-independent way of showing
# dialogs to the user when the main Foliage window is not available.
# (E.g., before the main GUI loop is started.)

qtapp = QApplication([''])

def tell_success(text):
    '''Wrapper around put_success(...) that also formats markdown.'''
    log(antiformat(text))
    put_success(put_markdown(text).style('margin-left: -3px'))


def tell_warning(text):
    '''Wrapper around put_warning(...) that also formats markdown.'''
    log(antiformat(text))
    put_warning(put_markdown(text).style('margin-left: -3px'))


def tell_failure(text):
    '''Wrapper around put_failure(...) that also formats markdown.'''
    log(antiformat(text))
    put_error(put_markdown(text).style('margin-left: -3px'))


def note_info(text):
    '''Show an informational toast message.'''
    log(antiformat(text))
    if os.environ.get('FOLIAGE_GUI_STARTED', 'False') == 'True':
        toast(text, color = 'green')
    elif inside_pyinstaller_app():
        # We don't print info-level msgs in this case.
        pass
    else:
        from rich import print
        print('[green]' + text + '[/]')


def note_warn(text):
    '''Show a warning toast message.'''
    log(antiformat(text))
    if os.environ.get('FOLIAGE_GUI_STARTED', 'False') == 'True':
        toast(text, color = 'warn')
    elif inside_pyinstaller_app():
        # Close the PyInstaller app splash screen if it's still visible.
        close_splash_screen()
        title = 'Foliage' if os.name == 'nt' else 'Foliage warning'
        QMessageBox.warning(None, title, 'Warning: ' + text)
    else:
        from rich import print
        width = 79 if len(text) > 75 else (len(text) + 4)
        print(Panel(text, style = Style.parse('yellow'), width = width))


def note_error(text):
    '''Show an error toast message.'''
    log(antiformat(text))
    if os.environ.get('FOLIAGE_GUI_STARTED', 'False') == 'True':
        toast(text, color = 'error')
    elif inside_pyinstaller_app():
        # Close the PyInstaller app splash screen if it's still visible.
        close_splash_screen()
        title = 'Foliage' if os.name == 'nt' else 'Foliage error'
        QMessageBox.critical(None, title, 'Error: ' + text)
    else:
        from rich import print
        width = 79 if len(text) > 75 else (len(text) + 4)
        print(Panel(text, style = Style.parse('red'), width = width))



#!/usr/bin/env python3
# =============================================================================
# @file    setup.py
# @brief   Installation setup file
# @created 2021-10-16
# @license Please see the file named LICENSE in the project directory
# @website https://github.com/caltechlibrary/foliage
#
# Note: configuration metadata is maintained in setup.cfg.  This file exists
# primarily to hook in setup.cfg and requirements.txt.
# =============================================================================

import os
from   os.path import exists, join, abspath, dirname
from   setuptools import setup

here = abspath(dirname(__file__))

requirements = []
if exists(join(here, 'requirements.txt')):
    with open(join(here, 'requirements.txt')) as f:
        requirements = f.read().rstrip().splitlines()

setup(
    setup_requires = ['wheel'],
    install_requires = requirements,
)



'''
system-widget.py: class for creating a taskbar or menubar widget

A disadvantage of using a PyWebIO-based approach is that there is no obvious
indication to the user whether Foliage is running, aside from looking at the
web page that serves as Foliage's interface.  This is not great: users
typically have a lot of browser windows and tabs open, and could lose track
of the one for Foliage. If the user switches away from the web page or
minimizes the window or does anything that might cause them to lose sight of
it, then they may forget about it altogether or be confused about what's
happening.

A traditional desktop application on Windows or macOS would have an
indication, such as an icon in the Windows taskbar or in the macOS Dock, or
alternatively, an icon in the macOS status bar, showing that the application
is still running.  This widget would be tightly coupled with the application
itself, such that exiting from the widget would exit Foliage too and exiting
Foliage would close the widget.

There is no built-in capability in PyWebIO itself to do this, so I had to
develop a novel approach.  It turned out to be much, much harder than
expected.

The biggest obstacle to doing it is how GUI libraries expect event loops to
be implemented, and on macOS, a related problem involving GUI threads.
Libraries like PyQt require you to execute a blocking call (in the case of
PyQt, it's app.exec_()) that starts the user interaction, but here in
Foliage, we have to start a different event loop controlled by PyWebIO.  So,
we can't use app.exec_() in the main thread.  This leads to the need to use
separate threads, which then leads to more problems:

1) On macOS, you can't run GUI event loops anywhere except in the main
   application thread, but as mentioned above, we can't do that because our
   main application thread is started by PyWebIO.  Doing so will result in an
   error from macOS about operations not being allowed outside the main
   thread.  In the end, on macOS, I couldn't find a way to make things work
   with PyQt at all: even trying to run a subprocess for the PyQt widget
   resulted in errors about not being in the main application thread.
   Non-PyQt solutions failed for similar reasons.  Basically, no GUI toolkit
   worked in a subthread.  (Thankfully, on Windows, this problem doesn't seem
   to exist, and a PyQt solution is possible.)  I briefly looked at PyWebIO's
   coroutines scheme as an alternative to its default subthread approach, but
   it wasn't immediately obvious how it would help solve the problem that we
   can't call app.exec_() from the Foliage main thread.

   The solution that I finally hit upon for macOS is to bundle a separate
   non-Python application to create a widget, and then using a subprocess (not
   a thread) to start that widget program.  This avoids the threading problem,
   at the cost of requiring a scheme for communicating with a separate process
   for the widget.

2) On Windows, where a subthread *can* be used for the PyQt widget, it was
   another challenge to figure out how to communicate to the main Foliage app
   when the user chooses to quit the widget.  It would have been simplest to
   have the widget code call our quit_app() directly.  It turns out that a
   subthread cannot issue PyWebIO calls because PyWebIO calls must be
   executed from the thread that started PyWebIO.  If you try, you don't get
   the kind of system errors you get when trying to use GUI event loops in a
   subthread, but the actions have no effect.  The widget function also can't
   throw an exception (e.g., SystemExit would have been an appropriate
   exception to throw for this purpose) because the exception won't end up in
   the main thread.  Thus, if you manage to get a GUI widget working in a
   subthread, you now face the problem of communicating to the parent thread
   that the user has selected "Quit" from the widget.

   After considerable time and many rabbit holes, I hit on the approach of
   (1) making the quit button in the PyWebIO UI use a PyWebIO pin object, so
   that testing for the quit button event can be done in the main event loop;
   (2) using a structured object (specifically a dict) containing a field
   value that is set inside the widget; (3) holding on to this structured
   object outside the widget, in the main thread, so that the main thread can
   test the value; and (4) using a timeout on the PyWebIO wait in the main
   foliage() "while True" loop, so the main thread can periodically test the
   value of that structured object to learn the current value set by the
   widget.  (See the end of the function foliage_page() in __main__.py.)

3) On Windows, the scheme worked.  On macOS, it worked on the command line
   when running "python3 -m foliage" or the equivalent, but ... failed in a
   PyInstaller-built application.  The problem is that PyInstaller does not
   bundle a Python interpreter, so running a separate Python script becomes
   impossible.  You don't want to make the assumption that the user's machine
   will have a Python installed.  After going down a number of rabbit holes,
   I finally hit on a working solution: use a separate, statically-linked
   program that is run by Foliage as a subprocess.  The binary can be bundled
   in the PyInstaller-built application so that Foliage knows where to find
   it.  The Go language provides an especially attractive solution because the
   binaries are very easy to build and statically-linked by default, and as
   luck would have it, a system tray package already existed and I was able to
   use it as a starting point (https://github.com/getlantern/systray).  The
   code for this widget written in Go is in data/macos-systray-widget/.


Copyright
---------

Copyright (c) 2021-2022 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

from   commonpy.interrupt import wait
import os
from   os.path import exists, dirname, join, basename, abspath, realpath, isdir
from   sidetrack import log
import sys


class SystemWidget():
    '''Encapsulate the control of a taskbar/system tray widget

    The approach taken here involves 2 separate systems:

    On Windows: the code for the widget is in this file, in the method
       start_windows_widget(). It  uses PyQt and runs the PyQt event loop
       in a subthread.  It sets the variable self.widget_info to a structured
       object (a dict), and sets a value inside that object when the PyQt
       event loop ends.  The PyQt event loop ends when the user selects "Quit"
       from the widget menu.  The use of a structured object (and not a simple
       Boolean variable) makes it possible to check the value outside the PyQt
       event loop, in the method running().

    On macOS: the widget is implemented as a separate program altogether.  The
       code is in the subdirectory data/macos-systray-widget/.  The method
       start_macos_widget() runs the separate program as a subprocess.  The
       method running() tests whether the process is still running; if it's
       not, it means either the user quit the widget using the "Quit" menu
       or else the widget was killed somehow externally to Foliage.
    '''

    def __init__(self):
        log('creating system widget')
        self.widget_info = None
        self.widget_process = None
        self.widget_thread = None
        if sys.platform.startswith('darwin'):
            self.start_macos_widget()
        else:
            self.start_windows_widget()


    def running(self):
        '''Return True if the taskbar/system tray widget is still running.

        In the Windows version of the widget (which uses PyQt GUI elements),
        a subthread runs the PyQt event loop.  The user can select the "Quit"
        menu item to indicate they want to exit Foliage.  If the PyQt event
        loop is not running, this method returns False.

        In the macOS version of the widget (which is a completely separate
        program written in Go), a subprocess runs the widget program.  The
        user can select the "Quit" menu item to exit the widget.  If the
        widget process is not running, this method returns False.
        '''
        if sys.platform.startswith('darwin'):
            return self.widget_process and (self.widget_process.poll() is None)
        else:
            return self.widget_info and self.widget_info['running']


    def stop(self):
        '''Stop the widget.

        On Windows, this does not actually need to do anything, because when
        the main Foliage application process exits then Python will kill
        daemon threads automatically.

        On macOS, this performs a process kill() on the subprocess, and if
        that fails, it sends a SIGTERM to the process.
        '''
        if not self.running():
            log('stop called for system widget but it is no longer running')
            return
        if self.widget_process:
            log('killing macos widget process')
            self.widget_process.kill()
            try:
                out, err = self.widget_process.communicate(timeout = 1)
                log('widget process output: ' + str(out))
            except:
                import signal
                log('sending SIGTERM to widget process')
                self.widget_process.send_signal(signal.SIGTERM)
            self.widget_process = None
        elif self.widget_info and self.widget_info['running']:
            # Nothing to do; it will get killed when Foliage exits.
            log('letting Windows widget get terminated normally on quit')


    def start_macos_widget(self):
        '''Start the Foliage system tray widget on macOS.'''
        import subprocess
        data_dir = realpath(join(dirname(__file__), 'data'))
        widget = join(data_dir, 'macos-systray-widget', 'macos-systray-widget')
        if exists(widget):
            log('starting macos systray widget: ' + widget)
            self.widget_process = subprocess.Popen(widget)
        else:
            log('macos widget binary is not at the expected path')


    def start_windows_widget(self):
        '''Start the taskbar widget on Windows.'''
        # We use a structured type, not a simple Boolean, because the value
        # needs to be set inside a thread but we need a handle on it from
        # outside the thread outside.  This next variable declaration puts it
        # in the lexical scope of the current function, such that the
        # show_widget() function below can access it.  Then, at the very end
        # of the current function, we set self.widget_info to point to this.
        # The indirection is needed because show_widget() can't get access
        # self.widget_info directly, because of scoping issues.
        widget_info = {'running': True}

        # The taskbar widget is implemented using PyQt and runs in a subthread.
        def show_widget():
            from PyQt5 import QtGui, QtWidgets, QtCore
            from PyQt5.QtCore import Qt

            log('creating Qt app for producing taskbar icon')
            app = QtWidgets.QApplication([])
            icon = QtGui.QIcon()
            data_dir = join(dirname(__file__), 'data')
            log('reading widget icons from ' + data_dir)
            icon.addFile(join(data_dir, 'foliage-icon-256x256.png'), QtCore.QSize(256,256))
            icon.addFile(join(data_dir, 'foliage-icon-128x128.png'), QtCore.QSize(128,128))
            icon.addFile(join(data_dir, 'foliage-icon-64x64.png'),   QtCore.QSize(64,64))
            app.setWindowIcon(icon)
            mw = QtWidgets.QMainWindow()
            mw.setWindowIcon(icon)
            mw.setWindowTitle('Foliage')
            mw.setWindowFlags(Qt.CustomizeWindowHint | Qt.WindowMinimizeButtonHint)
            mw.showMinimized()

            nonlocal widget_info
            log('starting windows taskbar widget')
            # The following call will block until the widget is exited (by
            # the user right-clicking on the widget and selecting "exit").
            app.exec_()

            # If the user right-clicks on the taskbar widget & chooses exit,
            # app.exec_() exits and we continue execution here.  We proceed to
            # set the value inside widget_info to False.
            log('taskbar widget returned from exec_()')
            widget_info['running'] = False

            # The main Foliage loop tests the running state (via the method
            # running() on our parent object) on a 1-second polling interval.
            # Here we have to wait longer than 1 second, to give the main
            # loop enough time to detect that the running state has changed
            # and communicate to the browser window to close itself.  If we
            # don't pause long enough here, then something bad happens: the
            # ending of the PyQt thread results in Foliage being killed
            # before the main loop has a chance to perform an orderly quit
            # (and specifically, before it has time to run JavaScript code in
            # the browser window to close the window). If the Foliage window
            # is left on the screen but the Foliage process has quit, it's
            # very confusing for users, so we want to avoid that.  I haven't
            # been able to figure out how to prevent the PyQt thread exit
            # from killing all of Foliage; I think the problem lies in how
            # PyQt uses signals to communicate events, but despite various
            # attempts to ignore signals in the main thread on Windows, I've
            # been unable to prevent it from happening.  (Python signals on
            # Windows are known to be problematic; see the following very
            # informative posting by Eryk Sun on 2016-03-04 on Stack
            # Overflow: https://stackoverflow.com/a/35792192/743730) The only
            # solution I have found so far is to make *this* code pause
            # longer than the 1-sec polling interval of the main loop, to
            # give the main loop enough time to do an orderly exit.  That is
            # the reason for the next line.
            wait(2)

        from threading import Thread
        log(f'starting taskbar icon widget in a subthread')
        thread = Thread(target = show_widget, daemon = True, args = ())
        thread.start()
        # Note we never join() the thread, b/c that would block.  We start the
        # widget thread & return so caller can proceed to its own event loop.

        self.widget_info = widget_info
